import java.util.*;

public class AgentKnowledge {
	int playerId, nPlayers, nFamilies, nGenerations;
	ArrayList<ArrayList <Integer>> kGroups = new ArrayList<ArrayList<Integer>>();
	HashSet<ArrayList <Integer>> cBelief = new HashSet<ArrayList<Integer>>(nFamilies);
	HashSet<ArrayList <Integer>> correct = new HashSet<ArrayList<Integer>>(nFamilies);
	HashSet<OlderThan> kGeneration;

	
	AgentKnowledge (int nF, int nG, int p) {
		playerId = p;
		nFamilies = nF;
		nGenerations = nG;
		nPlayers = nFamilies*nGenerations;
		kGeneration = new HashSet<OlderThan>();

		HashSet<Integer> e = new HashSet<Integer>(3);
		for (int i=0; i<nPlayers-2; i++) {
			e.add(i);
			for(int j=i+1; j<nPlayers-1; j++) {
				e.add(j);
				for (int k=j+1; k<nPlayers; k++) {
					e.add(k);
					HashSet<Integer> clone = (HashSet<Integer>) e.clone();
					kGroups.addAll(createOrdered(clone));
					e.remove(k);
				}
				e.remove(j);
			}
			e.remove(i);
		}
		
		createPossibleBeliefs();
	}
	
	public void printBelief() {
		Iterator<ArrayList<Integer>> itr = cBelief.iterator();
		
		while (itr.hasNext()) {
			System.out.print(itr.next() + " ");
		}
		System.out.println("");
	}
	
	public Stack<ArrayList <Integer>> createOrdered (HashSet<Integer> set) {
		//System.out.println(set + "");
		
		ArrayList<Integer> e = new ArrayList<Integer>(3);
		ArrayList<Integer> f = new ArrayList<Integer>(3);
		ArrayList<Integer> g = new ArrayList<Integer>(3);
		ArrayList<Integer> h = new ArrayList<Integer>(3);
		ArrayList<Integer> m = new ArrayList<Integer>(3);
		ArrayList<Integer> n = new ArrayList<Integer>(3);
		Stack<ArrayList<Integer>> ordered = new Stack<ArrayList<Integer>>();

		e.addAll(set);

		f.add(e.get(0));
		f.add(e.get(2));
		f.add(e.get(1));

		g.add(e.get(1));
		g.add(e.get(0));
		g.add(e.get(2));

		h.add(e.get(1));
		h.add(e.get(2));
		h.add(e.get(0));

		m.add(e.get(2));
		m.add(e.get(1));
		m.add(e.get(0));

		n.add(e.get(2));
		n.add(e.get(0));
		n.add(e.get(1));

		ordered.push(e);
		ordered.push(f);
		ordered.push(g);
		ordered.push(h);
		ordered.push(m);
		ordered.push(n);

		return ordered;
	}
	
	public void updateFamily(int p1, int p2, boolean family) {
		//System.out.println("updateFF: " + p1 + " " + p2 + " " + family);
		HashSet<Integer> e = new HashSet<Integer>();
		if (!family) { // Victim and playerId not in same team
			e.add(p1);
			e.add(p2);
			for (int i=0; i<nPlayers; i++) { // Eigenlijk een beetje vies, maar zo zijn wij.
				if (i==p1 || i==p2) continue;
				e.add(i);
 				Stack<ArrayList<Integer>> s = createOrdered(e);
				if (kGroups.removeAll(s) && playerId==0) {
					//System.out.println(e);
				}
				e.remove(i);
			}
		} else { // Victim and playerID in same team
			e.add(p1);
			for (int i=0; i<nPlayers;i++) {
				if (i==p1 || i==p2) continue;
				e.add(i);
				for (int j=i+1; j<nPlayers;j++) {
					if (j==p1 || j==p2) continue;
					e.add(j);
	 				Stack<ArrayList<Integer>> s = createOrdered(e);
					if (kGroups.removeAll(s) && playerId==0) {
						//System.out.println(e);
					}
					e.remove(j);
				}
				e.remove(i);
			}
			e.remove(p1);
			
			e.add(p2);
			for (int i=0; i<nPlayers;i++) {
				if (i==p1 || i==p2) continue;
				e.add(i);
				for (int j=i+1; j<nPlayers;j++) {
					if (j==p1 || j==p2) continue;
					e.add(j);
	 				Stack<ArrayList<Integer>> s = createOrdered(e);
					if (kGroups.removeAll(s) && playerId==0) {
						//System.out.println(e);
					}
					e.remove(j);
				}
				e.remove(i);
			}
			e.remove(p2);
		}
	}
	
	public void updateGeneration(int p1, int p2, int goodAnswer) {
		kGeneration.add(new OlderThan(p1, p2, goodAnswer));
	}
	
	public boolean isGenerationConsistent(HashSet<ArrayList<Integer>> belief) {
		boolean consistent = true;
		
		Iterator itr = kGeneration.iterator();
		while (itr.hasNext()) {
			if (!((OlderThan) itr.next()).isConsistent(belief)) consistent=false;
		}
		return consistent;
	}
	
	public void createPossibleBeliefs () {
		HashSet<Integer> e = new HashSet<Integer>();
		HashSet<Integer> f = new HashSet<Integer>();
		HashSet<Integer> g = new HashSet<Integer>();
		int c=0;
		boolean fetchNewBelief;
		//System.out.println("kGroups size: " + kGroups.size());
		
		// Check if current belief is still consistent
		if (kGroups.containsAll(cBelief) && cBelief.size()==nFamilies && isGenerationConsistent(cBelief)){
			fetchNewBelief = false;
		} else {
			fetchNewBelief = true;
		}
			
		//System.out.println("fetch belief? " + fetchNewBelief);
		
		Collections.shuffle(kGroups);
		
		if (fetchNewBelief) {
			for (int i=0; i<kGroups.size(); i++) {
				e.addAll(kGroups.get(i));
				//System.out.println(e.toString());
				
				for (int j=i+1; j<kGroups.size(); j++) {
					f = (HashSet<Integer>) e.clone();
					f.addAll(kGroups.get(j));
					if (f.size()<6)	continue;
						
						for (int l=j+1; l<kGroups.size(); l++) {
							g = (HashSet<Integer>) f.clone();
							g.addAll(kGroups.get(l));
							if (g.size()<9)	continue;
							
							c++;
							
							if (fetchNewBelief) {
								cBelief.clear();
								cBelief.add(kGroups.get(i));
								cBelief.add(kGroups.get(j));
								cBelief.add(kGroups.get(l));
								if (isGenerationConsistent(cBelief)) fetchNewBelief = false;
							}
							
							g.removeAll(kGroups.get(l));
							if (!fetchNewBelief) break;
					}
					f.removeAll(kGroups.get(j));
					if (!fetchNewBelief) break;
				}
				e.removeAll(kGroups.get(i));
				if (!fetchNewBelief) break;
			}
			//System.out.println("Total different belief states: " + c);
		}
		
		/*
		if (playerId==0) {
			Iterator itr = cBelief.iterator();
			while (itr.hasNext()) {
				System.out.print(itr.next() + " ");
			}
			System.out.println("");
		}*/
	}

	public double familyAnswer(int player1, int player2) {
		HashSet<ArrayList <Integer>> temp = new HashSet<ArrayList<Integer>>(nFamilies);
		HashSet<Integer> e = new HashSet<Integer>();
		HashSet<Integer> f = new HashSet<Integer>();
		HashSet<Integer> g = new HashSet<Integer>();
		
		int totalStates=0, sameTeam=0;
		
		for (int i=0; i<kGroups.size(); i++) {
			e.addAll(kGroups.get(i));
			//System.out.println(e.toString());
			
			for (int j=i+1; j<kGroups.size(); j++) {
				f = (HashSet<Integer>) e.clone();
				f.addAll(kGroups.get(j));
				if (f.size()<6)	continue;
					
					for (int l=j+1; l<kGroups.size(); l++) {
						g = (HashSet<Integer>) f.clone();
						g.addAll(kGroups.get(l));
						if (g.size()<9)	continue;
						
						totalStates++;
						
						int choice = 0;
						
						temp.clear();
						temp.add(kGroups.get(i));
						temp.add(kGroups.get(j));
						temp.add(kGroups.get(l));
						
						Iterator itr = temp.iterator();
						while (itr.hasNext()) {
							ArrayList<Integer> current = (ArrayList<Integer>) itr.next();
							if (current.contains(player1) && current.contains(player2))
								choice = 1;
						}
						
						if (isGenerationConsistent(cBelief) && choice==1) sameTeam++;
						
						g.removeAll(kGroups.get(l));
				}
				f.removeAll(kGroups.get(j));
			}
			e.removeAll(kGroups.get(i));
		}
		double p = ((double)sameTeam/(double)totalStates);
		return (p);
	}
	
	
}